#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <iostream>
#include <string>

#include "argumentsettings.h"
#include "config.h"

#define CONF_HOME_DIR ".gtkevemon"

/* Default configuration is always loaded. The application
 * can rely on these keys to be there all the time even if
 * the values are not in the configuration file. */
char const* default_config =
    "[accounts]\n"
    "[characters]\n"
    "[evetime]\n"
    "  valid = false\n"
    "  difference = 0\n"
    "  time_format = %Y-%m-%d %H:%M:%S\n"
    "  time_short_format = %m-%d %H:%M\n"
    "[network]\n"
    "  use_proxy = false\n"
    "  proxy_address = \n"
    "  proxy_port = 80\n"
    "[notifications]\n"
    "  show_popup_dialog = true\n"
    "  show_tray_icon = false\n"
    "  show_info_bar = true\n"
    "  exec_handler = false\n"
    "  exec_command = mail your@mail.org -s \"$CHAR finished $SKILL $LEVEL\"\n"
    "  exec_data = Congratulations, $CHAR finished $SKILL $LEVEL$NL$NLStatistics about the skill:$NLStart SP: $STARTSP$NLDestination SP: $DESTSP$NLAcquired SP: $NEWSP$NLSP/h: $SPPH$NLTotal training time: $TRAINTIME$NL$NLThis message was generated by GtkEveMon, $LOCALTIME\n"
    "  minimum_sp = 16000\n"
    "[planner]\n"
    "  columns_format = +0 +1 +2 -3 -4 -5 -6 -7\n"
    "  current_plan = \n"
    "[servermonitor]\n"
    "[settings]\n"
    "  auto_update_sheets = true\n"
    "  detailed_tray_tooltip = true\n"
    "  eve_command = eve\n"
    "  eve_command_2 = \n"
    "  eve_command_3 = \n"
    "  eve_command_4 = \n"
    "  eve_command_5 = \n"
    "  minimize_on_close = false\n"
    "  startup_servercheck = true\n"
    "  tray_usage = never\n"
    "  verbose_wintitle = true\n"
    "[versionchecker]\n"
    "  enabled = true\n"
    "  raise_updater = true\n"
    "[versionchecker.datafiles]\n"
    "  CertificateTree.xml.gz = \n"
    "  ItemTree.xml.gz = \n"
    "  SkillTree.xml.gz = \n"
    "[versionchecker.last_seen]\n";

/* The initial configuration is loaded once if the configuration
 * file is created for the first time. Thus it initializes the
 * configuration but the values can be deleted from the config. */
char const* initial_config =
    "[servermonitor]\n"
    "  0_Tranquility = 87.237.38.200\n"
    "  1_Serenity = 61.129.54.100\n"
    "  2_Singularity = 87.237.38.50\n";

/* ---------------------------------------------------------------- */

Conf Config::conf;
std::string Config::conf_dir;
std::string Config::filename;

/* ---------------------------------------------------------------- */

void
Config::init_defaults (void)
{
  Config::conf.add_from_string(default_config);
}

/* ---------------------------------------------------------------- */

void
Config::init_user_config (void)
{
  /* The first part here determines the user's config directory.
   * If there is no home directory or any other error occures,
   * the current directory or even worse, /tmp is used. */
  std::string user_conf_dir;
  if (ArgumentSettings::config_dir.empty())
  {
    uid_t user_id = ::geteuid();
    struct passwd* user_info = ::getpwuid(user_id);

    if (user_info == 0 || user_info->pw_dir == 0)
    {
      std::cout << "Warning: Couldn't determine home directry!" << std::endl;
      char buffer[512];
      char* ret = ::getcwd(buffer, 512);
      if (ret != 0)
        user_conf_dir = buffer;
      else
      {
        std::cout << "Warning: Couldn't even determine CWD!" << std::endl;
        user_conf_dir = "/tmp";
      }
    }
    else
    {
      user_conf_dir = user_info->pw_dir;
      user_conf_dir += "/";
      user_conf_dir += CONF_HOME_DIR;
    }
  }
  else
  {
    user_conf_dir = ArgumentSettings::config_dir;
  }

  //std::cout << "Resolved config directory to: " << user_conf_dir << std::endl;

  /* Check if the GtkEveMon directory exists in the home directry. */
  int dir_exists = ::access(user_conf_dir.c_str(), F_OK);

  if (dir_exists < 0)
  {
    /* Directory does not exists. Create it. */
    std::cout << "Creating config directory: " << user_conf_dir << std::endl;
    int ret = ::mkdir(user_conf_dir.c_str(), S_IRWXU);
    if (ret < 0)
    {
      std::cout << "Error: Couldn't create the config directory!" << std::endl;
      std::cout << "Error: Falling back to /tmp" << std::endl;
      user_conf_dir = "/tmp";
    }
  }

  /* Set the filename, make it official. */
  Config::conf_dir = user_conf_dir;
  Config::filename = user_conf_dir + "/gtkevemon.conf";

  /* Check if the config file is in-place. If not, dump the current
   * default configuration to file. */
  int conf_exists = ::access(Config::filename.c_str(), F_OK);
  if (conf_exists < 0)
  {
    std::cout << "Creating initial config file: gtkevemon.conf" << std::endl;
    Config::conf.add_from_string(initial_config);
    Config::conf.to_file(Config::filename);
  }
  else
  {
    /* Now read the config file. If it's not there,
     * it will be created when GtkEveMon exits. */
    Config::conf.add_from_file(Config::filename);
  }
}

/* ---------------------------------------------------------------- */

void
Config::setup_http (AsyncHttp* fetcher)
{
  fetcher->set_agent("GtkEveMon");

  ConfSectionPtr section = Config::conf.get_section("network");
  if (section->get_value("use_proxy")->get_bool())
  {
    fetcher->set_proxy(section->get_value("proxy_address")->get_string(),
        (uint16_t)section->get_value("proxy_port")->get_int());
  }
}
